import copyreg
import dataclasses
import hashlib
import io
import json
import pickle
import torch


def _reduce_tensor(tensor):
    """__reduce__ for torch.Tensor.

    Since pickle.dumps(tensor) is not deterministic, we convert torch.Tensor to numpy array.
    """
    return torch.Tensor, (tensor.cpu().numpy(),)


class HashPickler(pickle.Pickler):
    """Pickler to calculate hash.

    Note that it's not guaranteed that the loads(dumps(obj)) will create the same object.
    """
    dispatch_table = copyreg.dispatch_table.copy()
    dispatch_table[torch.Tensor] = _reduce_tensor


class HashGenerator:
    """Calculate hash for the given object.

    Note that we cannot simply hashlib.sha1(pickle.dumps(obj)) since some of the contents might be cached on remote. For nested objects, we calculate hash for each child element recursively.
    """
    @classmethod
    def calculate_hash(cls, value, context=None):
        from .variable import Variable

        def get_hash(value):
            if isinstance(value, dict):
                value = json.dumps({k: get_hash(v) for k, v in sorted(value.items())}).encode('utf-8')
            elif isinstance(value, list):
                value = json.dumps([get_hash(v) for v in value]).encode('utf-8')
            elif dataclasses.is_dataclass(value):
                value = json.dumps({k: get_hash(v) for k, v in sorted(dataclasses.asdict(value).items())}).encode('utf-8')
            elif isinstance(value, Variable):
                return value.get_hash(context)
            elif hasattr(value, '__getstate__'):
                return get_hash(value.__getstate__())
            else:
                f = io.BytesIO()
                p = HashPickler(f)
                p.dump(value)
                value = f.getbuffer()

            return hashlib.sha1(value).hexdigest()

        return get_hash(value)



import dataclasses
import sys
import unittest
from typing import Dict, List, Optional
from irisml.core import Context, TaskBase, TaskDescription
from irisml.core.task import Task


class TestTask(unittest.TestCase):
    def test_load_config(self):
        @dataclasses.dataclass
        class ChildConfig:
            int_var: int

        @dataclasses.dataclass
        class Config:
            int_var: int
            str_var: str
            optional_int_var: Optional[int]
            int_list_var: List[int]
            str_dict_var: Dict[str, str]
            nested: ChildConfig
            int_default_var: int = 42

        result = Task._load_config(Config, {'int_var': 12, 'str_var': '12', 'optional_int_var': 24, 'int_list_var': [1, 2, 3], 'str_dict_var': {'k': 'v', 'k2': 'v2'}, 'nested': {'int_var': 123}})
        self.assertEqual(result.nested, ChildConfig(int_var=123))
        self.assertEqual(result.int_var, 12)
        self.assertEqual(result.str_var, '12')
        self.assertEqual(result.optional_int_var, 24)
        self.assertEqual(result.int_list_var, [1, 2, 3])
        self.assertEqual(result.str_dict_var, {'k': 'v', 'k2': 'v2'})
        self.assertEqual(result.int_default_var, 42)

    def test_validate(self):
        class CustomTask:
            class Task(TaskBase):
                @dataclasses.dataclass
                class Outputs:
                    int_value: int

        class EmptyClass:
            pass

        with unittest.mock.patch.dict(sys.modules):
            sys.modules['irisml.tasks.custom_task'] = CustomTask
            task = Task(TaskDescription.from_dict({'task': 'custom_task'}))
            with self.assertRaises(RuntimeError):
                task.load_module()

            CustomTask.Task.Outputs = EmptyClass
            task = Task(TaskDescription.from_dict({'task': 'custom_task'}))
            with self.assertRaises(RuntimeError):
                task.load_module()

    def test_inputs_env(self):
        """Environment variable in inputs should be casted to the expected type."""
        task_description = TaskDescription.from_dict({'task': 'custom_task', 'inputs': {'input0': '$env.ENV_INPUT'}})

        class CustomTask:
            class Task(TaskBase):
                @dataclasses.dataclass
                class Inputs:
                    input0: int

                def execute(self, inputs):
                    if not isinstance(inputs.input0, int):
                        raise RuntimeError(f"Input type is {type(inputs.input0)}")
                    return self.Outputs()

        context = Context({'ENV_INPUT': '12345'})
        with unittest.mock.patch.dict(sys.modules):
            sys.modules['irisml.tasks.custom_task'] = CustomTask
            task = Task(task_description)
            task.load_module()
            task.execute(context)



from .cache_manager import CachedOutputs, HashGenerator


def replace_variables(value):
    """Replace a string '$env' and '$outputs' in the given object with Variable instances."""
    if isinstance(value, dict):
        return {k: replace_variables(v) for k, v in value.items()}
    elif isinstance(value, list):
        return [replace_variables(v) for v in value]
    elif isinstance(value, str) and value[0] == '$':
        if value.startswith('$env.'):
            return EnvironmentVariable(value)
        elif value.startswith('$output.'):
            return OutputVariable(value)
        else:
            raise ValueError(f"Unknown variable type: {value}")
    else:
        return value


class Variable:
    """Base class for Variables.

    In a task description, a string with '$' prefix is considered a variable.
    Currently we have two variable types: $env and $output. $env is Environment Variables that can be set to Context. $output is output objests from previous tasks.

    """
    def __init__(self, name):
        self._var_str = name
        self._expected_type = str

    @property
    def expected_type(self):
        return self._expected_type

    @expected_type.setter
    def expected_type(self, value):
        self._expected_type = value

    def __eq__(self, other):
        return type(self) is type(other) and self._var_str == other._var_str

    def __hash__(self):
        return hash(self._var_str)

    def resolve(self, context):
        """Returns the actual value this variable represents."""
        raise NotImplementedError

    def get_hash(self, context):
        """Get hash value for the value that is stored in this Variable."""
        return HashGenerator.calculate_hash(self.resolve(context))


class EnvironmentVariable(Variable):
    """A variable with $env prefix"""
    def __init__(self, name):
        super().__init__(name)
        parts = name.split('.')
        if len(parts) != 2 or parts[0] != '$env' or not parts[1].isupper():
            raise ValueError(f"Invalid environment variable name: {name}")

        self._name = parts[1]

    def resolve(self, context):
        return self._expected_type(context.get_environment_variable(self._name))

    def __str__(self):
        return f"EnvVar({self._name})"


class OutputVariable(Variable):
    """A variable with $output prefix. Outputs from a task."""
    def __init__(self, name):
        super().__init__(name)
        parts = name.split('.')
        if len(parts) != 3 or parts[0] != '$output' or not name.islower():
            raise ValueError(f"Invalid output variable name: {name}")

        self._name = parts[1]
        self._path = parts[2]

    def resolve(self, context):
        outputs = context.get_outputs(self._name)
        if not hasattr(outputs, self._path):
            raise ValueError(f"Output {self._name} doesn't have path {self._path}")
        return getattr(outputs, self._path)

    def get_hash(self, context):
        outputs = context.get_outputs(self._name)
        if isinstance(outputs, CachedOutputs):
            return outputs.get_hash(self._path)

        if not hasattr(outputs, self._path):
            raise ValueError(f"Output {self._name} doesn't have path {self._path}")
        value = getattr(outputs, self._path)
        return HashGenerator.calculate_hash(value, context)

    def __str__(self):
        return f"OutputVar({self._name}.{self._path})"



import unittest
from irisml.core.context import Context
from irisml.core.variable import EnvironmentVariable, OutputVariable


class TestContext(unittest.TestCase):
    def test_clone(self):
        context = Context({'env': 'value'})
        new_context = context.clone()
        self.assertEqual(new_context.get_environment_variable('env'), 'value')

        new_context.add_environment_variable('env2', 'value2')
        new_context.add_outputs('out', None)

        with self.assertRaises(ValueError):
            context.get_environment_variable('env2')

        with self.assertRaises(ValueError):
            context.get_outputs('out')

    def test_resolve(self):
        context = Context()
        self.assertEqual(context.resolve(123), 123)
        self.assertEqual(context.resolve('123'), '123')
        self.assertEqual(context.resolve([1, 2, 3]), [1, 2, 3])
        self.assertEqual(context.resolve([1, '2', 3]), [1, '2', 3])

        with self.assertRaises(ValueError):
            context.resolve(EnvironmentVariable('$env.E'))

        context = Context({'E': 'v'})
        self.assertEqual(context.resolve(EnvironmentVariable('$env.E')), 'v')
        self.assertEqual(context.resolve([3, EnvironmentVariable('$env.E')]), [3, 'v'])

        class DummyOutput:
            value = 3

        context.add_outputs('test_out', DummyOutput())
        self.assertEqual(context.resolve([123, OutputVariable('$output.test_out.value')]), [123, 3])



from irisml.core.context import Context
from irisml.core.description import JobDescription, TaskDescription
from irisml.core.task import Task
from irisml.core.task_base import TaskBase

__all__ = ['Context',
           'JobDescription', 'TaskDescription',
           'Task',
           'TaskBase']
